/**
 * Illustrates global memory accesses
 *
 * Lecture notes:
 * 1) Understand the global memory access pattern of transpose_naive. Can we improve it? 
 * 2) Profile the naive version. Have a look at "Global store efficiency"
 * 3) Profile sm version. Have a look at "Shared Memory Replay Overhead", which is an indicator for bank conflicts.
 * 4) Remove bank conflict
 */

#include <cuda_runtime.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "../misc/cuda_util.h"

#define BLOCKSIZE 128
#define NUM_BLOCKS 128


int main(int argc, char** argv)
{
   int i;
   int n= 1024*1024*10;
   if(argc > 1)
      n = atoi(argv[1]);

   int size = sizeof(float) *n;
   printf("Allocates %f MB\n", ((float)size)/1024/1024);


   //allocate host memory
   float *x_h;
   x_h = (float*) malloc(size);


   //init x and y with random numbers in [0,1]
   for(i=0;i < n; ++i){
      x_h[i] = ((float)rand())/RAND_MAX;
   }


   //allocate device memory
   float *x_d;

   /**********************************************
    * Normal copy 
    ***********************************************/
   HANDLE_ERROR( cudaMalloc(&x_d, size) );

   float time;
   cudaEvent_t start, stop; //used for timings
   HANDLE_ERROR( cudaEventCreate(&start) );
   HANDLE_ERROR( cudaEventCreate(&stop) );
   HANDLE_ERROR( cudaEventRecord(start, 0) ); 

   //copy data from host to device
   HANDLE_ERROR( cudaMemcpy(x_d, x_h, size, cudaMemcpyHostToDevice) ); 

   HANDLE_ERROR( cudaEventRecord(stop, 0) );
   HANDLE_ERROR( cudaEventSynchronize(stop) );
   HANDLE_ERROR( cudaEventElapsedTime(&time, start, stop) ); 
   printf("Normal bandwidth:  %5.3f GB/s \n", size*1000.0f/1024.0f/1024.0f/1024.0f/time);

   //free memory
   free(x_h);

   /**********************************************
    * Copy with page-locked memory
    ***********************************************/
   HANDLE_ERROR( cudaMallocHost(&x_h, size) ); //allocate page-locked memory

   HANDLE_ERROR( cudaEventCreate(&start) );
   HANDLE_ERROR( cudaEventCreate(&stop) );
   HANDLE_ERROR( cudaEventRecord(start, 0) ); 

   //copy data from host to device
   HANDLE_ERROR( cudaMemcpy(x_d, x_h, size, cudaMemcpyHostToDevice) ); 

   HANDLE_ERROR( cudaEventRecord(stop, 0) );
   HANDLE_ERROR( cudaEventSynchronize(stop) ); 
   HANDLE_ERROR( cudaEventElapsedTime(&time, start, stop) ); 
   printf("Page-locked bandwidth:  %5.3f GB/s \n", size*1000.0f/1024.0f/1024.0f/1024.0f/time);


   //free memory
   HANDLE_ERROR( cudaFreeHost(x_h) );
   HANDLE_ERROR( cudaFree(x_d) ) ;

   return 0;
}




